/* newtnipl.c
 * Newton-Interpolation
 * erstellt: Fre Nov 19 16:46:11 CET 1999
 * Autoren:  Mirko Belickin und Joachim Wagner
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include "interplt.h"
#include "namen.h"
#include "matrix.h"
#include "nmmatrix.h"
#include "newtnipl.h"

#ifndef PI
  #define PI (4*atan(1))
#endif

/* externes Klassenobjekt einbinden */

extern matrix nm_matrix;


/* Deklaration der Objekt-Methoden */

interpolation * newtoninterpolation_laden(FILE *datei, void *vater);
int newtoninterpolation_speichern(interpolation *m, FILE *datei);
interpolation * newtoninterpolation_kreieren(void *schablone, void *vater);
int newtoninterpolation_freigeben(interpolation *m, void *vater);
double newtoninterpolation_funktionswert(interpolation *m, double x);
char *newtoninterpolation_name(interpolation *m);
matrix * newtoninterpolation_imVektor(interpolation *m, int vektorNr);
double newtoninterpolation_linkeGrenze(interpolation *m);
double newtoninterpolation_rechteGrenze(interpolation *m);


/* Globales Klassenobjekt */

interpolationFunktionen newton_interpolationFunktionen = {
    newtoninterpolation_speichern,
    newtoninterpolation_laden,
    newtoninterpolation_freigeben,
    newtoninterpolation_kreieren,
    newtoninterpolation_funktionswert,
    newtoninterpolation_name,
    interpolation_klasse,
    newtoninterpolation_imVektor,
    newtoninterpolation_linkeGrenze,
    newtoninterpolation_rechteGrenze
};

interpolation newton_interpolation = {          /* das Klassenobjekt zur Klasse NM_interpolation */
    &newton_interpolationFunktionen,     
    NULL                      /* Klassenobjekt hat keine Daten */
};


/* private Methoden */

static newtoninterpolationsdaten *get_newtoninterpolationsdaten(interpolation *m);
static int lese_datum(FILE *datei, char *format, char * prefix, void *pdatum);


/* ------------------------------------------------------------------------ 
 * newton-interpolation Implementation
 */ 


interpolation * newtoninterpolation_laden(FILE *datei, void *vater)
{
  newtoninterpolationsschablone schablone;
  interpolation *m;

  schablone.name = interpolation_kopfzeile(datei, S_NEWTON_INTERPOLATION);

  /* Abmasse der interpolation feststellen */

  lese_datum(datei, "%d", "art =", &(schablone.stuetzstellenart));

  switch (schablone.stuetzstellenart) {
    case 0: schablone.x = nm_matrix.aktion->laden(datei, NULL);
            schablone.y = nm_matrix.aktion->laden(datei, NULL);
            break;
    case 1: case 2:
            lese_datum(datei, "%lf", "a =", &(schablone.a));
            lese_datum(datei, "%lf", "b =", &(schablone.b));
            /* lese_datum(datei, "%d", "n =", &(schablone.n)); */
            schablone.y = nm_matrix.aktion->laden(datei, NULL);
            break;
  }

  /* interpolation erzeugen */

  m = newton_interpolation.aktion->kreieren(&schablone, vater);

  /* Temporaere Daten wieder freigeben */
  if (schablone.stuetzstellenart==0)
    schablone.x->aktion->freigeben(schablone.x, NULL);
  schablone.y->aktion->freigeben(schablone.y, NULL);

  return m;
}

static int lese_datum(FILE *datei, char *format, char * prefix, void *pdatum)
{
  char eingabe[BUFSIZ];
  if (!fgets(eingabe,BUFSIZ,datei)) {
    fprintf(stderr,"Zeile %s nicht gefunden - EOF.\n", prefix);
    exit(1);
  }
  if (strncmp(eingabe, prefix,
               strlen( prefix))!=0) {
    fprintf(stderr,"%s... erwartet!\n", prefix);
    exit(1);
  }
  sscanf(&(eingabe[strlen(prefix)]), format, pdatum);
                               /* erstmal ohne Fehlerkontrolle */
  return 0;
}

int newtoninterpolation_speichern(interpolation *m, FILE *datei)
{
  newtoninterpolationsdaten *pm;
  pm = get_newtoninterpolationsdaten(m);
  fprintf(datei,S_NEWTON_INTERPOLATION ": %s\n", pm->name);
  fprintf(datei,"art = %i\n", pm->stuetzstellenart);
  switch (pm->stuetzstellenart) {
    case 0:
            pm->x->aktion->speichern(pm->x, datei);
            break;
    case 1: 
    case 2:
            fprintf(datei,"a = %lf\n", pm->a);
            fprintf(datei,"b = %lf\n", pm->b);
            break;
    default:
            fprintf(stderr, "Kann " S_NEWTON_INTERPOLATION " %s nicht "
            "speichern.\n", pm->name);
            exit(1);
  }
  pm->y->aktion->speichern(pm->y, datei);
  return 0;
}

interpolation * newtoninterpolation_kreieren(void *schablone, void *vater)
{
  newtoninterpolationsschablone *s = (newtoninterpolationsschablone *) schablone;
  interpolation *m;
  newtoninterpolationsdaten *pm;
  char *name;
  int i, k;
  nmmatrixschablone schablone2;

  if (!(m = (interpolation *)malloc(sizeof(interpolation)))) {
    fprintf(stderr,"Nicht genug Speicher fuer neue interpolation (Teil 1)!\n");
    exit(1);          
  }
  m->aktion = newton_interpolation.aktion;    /* moegliche Aktionen festlegen */
  if (!(m->daten = 
    (newtoninterpolationsdaten *)malloc(sizeof(newtoninterpolationsdaten)))) {
    fprintf(stderr,"Nicht genug Speicher fuer neue interpolation (Teil 2)!\n");
    exit(1);          
  }
  pm = get_newtoninterpolationsdaten(m);

  /* Name der interpolation festlegen */

  name = (s && s->name)?s->name:"unbenannt";
  if (!(pm->name = (char *)malloc( sizeof(char) * (1+strlen(name))))) {
    fprintf(stderr, "Nicht genug Speicher fuer " S_NEWTON_INTERPOLATION 
                    " %s.\n", name);
    exit(1);
  }
  strcpy(pm->name, name);

  /* Groesse der interpolation festlegen */

  if (!s || !(s->y)) {
    fprintf(stderr, "Stuetzwerte fehlen fuer" S_NEWTON_INTERPOLATION " %s.\n",
                    name);
    exit(1);
  }
  pm->n = s->y->aktion->hoehe(s->y) - 1;
  if (pm->n < 0) {
    fprintf(stderr,"Es wird mindestens ein Stuetzwert benoetigt.\n");
    exit(1);
  }
  schablone2.n = pm->n + 1;
  schablone2.m = 1;
  schablone2.init = 0;
  schablone2.name = "Stuetzstellen";
  pm->x = nm_matrix.aktion->kreieren(&schablone2, m);
  schablone2.name = "Stuetzwerte";
  pm->y = nm_matrix.aktion->kreieren(&schablone2, m);
  schablone2.name = "Koeffizienten";
  pm->c = nm_matrix.aktion->kreieren(&schablone2, m);

  switch(s->stuetzstellenart) {
    case 0:
          if (!(s->x)) {
            fprintf(stderr, "Stuetzstellen fehlen fuer" S_NEWTON_INTERPOLATION 
                            " %s.\n", name);
            exit(1);
          }
          if ((pm->n + 1) != (s->x->aktion->hoehe(s->x))) {
            fprintf(stderr, "Stuetzstellenanzahl/-werteanzahl inkompatibel " 
                            "fuer " S_NEWTON_INTERPOLATION " %s.\n", name);
            exit(1);
          }
          for (i=1; i <= (pm->n + 1); i++)
            pm->x->aktion->setzen(pm->x, i, 1,
              s->x->aktion->amIndex(s->x, i, 1));
          pm->stuetzstellenart = 0;
          break;

    case 1: /* aequidistante Stuetzstellen */
          if ((s->b - s->a) < EPSILON) {
            fprintf(stderr,"Intervall [%lf,%lf] numerisch zu klein oder "
                           "ungueltig.\n", s->a, s->b);
            exit(1);
          }
          for (i=0; i <= pm->n; i++)
            pm->x->aktion->setzen(pm->x, i+1, 1,
              (i * s->b + (pm->n - i) * s->a) / pm->n);   /* i-te Stuetz-   */
                                                          /* stelle, Formel */
                    /* umgeformt, so dass numerisch x(i) > b nicht auftritt */
          pm->stuetzstellenart = 1;
          pm->a = s->a;
          pm->b = s->b;
          break;
                
    case 2: /* Stuetzstellen nach Tschebyscheff-Polynom */
          if ((s->b - s->a) < EPSILON) {
            fprintf(stderr,"Intervall [%lf,%lf] numerisch zu klein oder "
                           "ungueltig.\n", s->a, s->b);
            exit(1);
          }
          for (i=0; i <= pm->n; i++)
            pm->x->aktion->setzen(pm->x, i+1, 1,
              (s->a + s->b)/2.0 + (s->a - s->b)*cos(i*PI/pm->n)/2.0);
          pm->stuetzstellenart = 2;
          pm->a = s->a;
          pm->b = s->b;
          break;
  }
  for (i=1; i <= (pm->n + 1); i++)
    pm->y->aktion->setzen(pm->y, i, 1,
      s->y->aktion->amIndex(s->y, i, 1));
 
  /* Koeffizienten bestimmen */
  for (i=0; i<= pm->n; i++) 
    pm->c->aktion->setzen(pm->c, i+1, 1,            /* c(i) = y(i) */
      pm->y->aktion->amIndex(pm->y, i+1, 1));
  for (k=1; k<= pm->n; k++)
    for (i= pm->n; i>=k; i--)
      pm->c->aktion->setzen(pm->c, i+1, 1,
        (pm->c->aktion->amIndex(pm->c, i+1, 1)
         - pm->c->aktion->amIndex(pm->c, i, 1))
        / (pm->x->aktion->amIndex(pm->x, i+1, 1)
           - pm->x->aktion->amIndex(pm->x, i+1-k, 1)));
  return m;
}

int newtoninterpolation_freigeben(interpolation *m, void *vater) 
{ 
  newtoninterpolationsdaten *pm;
  pm = get_newtoninterpolationsdaten(m);
  free(pm->name);
  pm->x->aktion->freigeben(pm->x, m);
  pm->y->aktion->freigeben(pm->y, m);
  pm->c->aktion->freigeben(pm->c, m);
  free(pm);
  free(m);
  return 0;              /* immer Erfolg */
}

static newtoninterpolationsdaten *get_newtoninterpolationsdaten(interpolation *m) 
{
  if ((!(m->daten))) /* sicher ist sicher ... */
    exit(1);                        /* Fehlermeldung fehlt noch */
  return m->daten;                  /* Daten der interpolation zurueckgeben */
}

char *newtoninterpolation_name(interpolation *m)
{
  newtoninterpolationsdaten *pm;
  if (!m->daten) 
    return S_NEWTON_INTERPOLATION;
  pm = get_newtoninterpolationsdaten(m);
  return pm->name; 
}

double newtoninterpolation_funktionswert(interpolation *m, double x)
{
  newtoninterpolationsdaten *pm;
  double p;
  int k;
  pm = get_newtoninterpolationsdaten(m);
  /* Horner-Schema */
  p = pm->c->aktion->amIndex(pm->c, pm->n + 1, 1);     /* p = c(n+1)       */
  for (k= pm->n; k>=1; k--) {
    p *= (x - pm->x->aktion->amIndex(pm->x, k, 1));    /* p = p * (x-x(k)) */
    p += pm->c->aktion->amIndex(pm->c, k, 1);          /*     + c(k)       */
  }
  return p;
}

matrix * newtoninterpolation_imVektor(interpolation *m, int vektorNr)
{
  newtoninterpolationsdaten *pm;
  pm = get_newtoninterpolationsdaten(m);
  switch (vektorNr) {
    case 0: return pm->x;
    case 1: return pm->y;
    case 2: return pm->c;
    default:
    fprintf(stderr,"Ungueltiger Vektor fuer Interpolationsproblem %s: (%i)\n",
                   pm->name, vektorNr);
    exit(1); 
  }
}

double newtoninterpolation_linkeGrenze(interpolation *m)
{
  newtoninterpolationsdaten *pm;
  double min, h;
  int i;
  int n;
  pm = get_newtoninterpolationsdaten(m);
  switch (pm->stuetzstellenart) {
    case 0:
    n = pm->x->aktion->hoehe(pm->x);
    min = pm->x->aktion->amIndex(pm->x, 1, 1);
    for (i=2; i<=n; i++) {
      h = pm->x->aktion->amIndex(pm->x, i, 1);
      if (h<min) min = h;
    }
    return min;
    case 1:
    case 2:
    return pm->a;
    break;
    default:
    fprintf(stderr,"Ungueltige Stuetzstellenart fuer Interpolationsproblem %s: (%i)\n",
                   pm->name, pm->stuetzstellenart);
    exit(1); 
  }
}

double newtoninterpolation_rechteGrenze(interpolation *m)
{
  newtoninterpolationsdaten *pm;
  double max, h;
  int i;
  int n;
  pm = get_newtoninterpolationsdaten(m);
  switch (pm->stuetzstellenart) {
    case 0:
    n = pm->x->aktion->hoehe(pm->x);
    max = pm->x->aktion->amIndex(pm->x, 1, 1);
    for (i=2; i<=n; i++) {
      h = pm->x->aktion->amIndex(pm->x, i, 1);
      if (h>max) max = h;
    }
    return max;
    case 1:
    case 2:
    return pm->b;
    break;
    default:
    fprintf(stderr,"Ungueltige Stuetzstellenart fuer Interpolationsproblem %s: (%i)\n",
                   pm->name, pm->stuetzstellenart);
    exit(1); 
  }
}

